@@ -1,3 +1,5 @@
+from bleak_retry_connector import close_stale_connections_by_address
+
 from .const import (
     ConnectionInfo,
     DoorStatus,
@@ -18,7 +20,7 @@ from .util import (
     unique_id_from_local_name_address,
 )
 
-__version__ = "2.1.13"
+__version__ = "2.4.1"
 
 __all__ = [
     "AuthError",
@@ -38,4 +40,5 @@ __all__ = [
     "local_name_is_unique",
     "YaleXSBLEDiscovery",
     "YaleXSBLEError",
+    "close_stale_connections_by_address",
 ]
