@@ -29,7 +29,7 @@ from .const import (
     LockStatus,
 )
 from .secure_session import SecureSession
-from .session import AuthError, DisconnectedError, Session
+from .session import AuthError, DisconnectedError, Session, YaleXSBLEError
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -103,6 +103,7 @@ class Lock:
             [Iterable[LockStatus | DoorStatus | BatteryState]], None
         ],
         info: LockInfo | None = None,
+        disconnect_callback: Callable[[], None] | None = None,
     ) -> None:
         self.ble_device_callback = ble_device_callback
         self.key = bytes.fromhex(keyString)
@@ -114,16 +115,23 @@ class Lock:
         self._lock = asyncio.Lock()
         self._lock_info = info
         self.client: BleakClientWithServiceCache | None = None
-        self._disconnected_event: asyncio.Event | None = None
         self._state_callback = state_callback
+        self._disconnected = False
+        self._disconnect_callback = disconnect_callback
+        self._disconnected_futures: set[asyncio.Future[None]] = set()
 
     def set_name(self, name: str) -> None:
         self.name = name
 
     def disconnected(self, *args: Any, **kwargs: Any) -> None:
         _LOGGER.debug("%s: Disconnected from lock callback", self.name)
-        assert self._disconnected_event is not None  # nosec
-        self._disconnected_event.set()
+        self._disconnected = True
+        for future in self._disconnected_futures:
+            if not future.done():
+                future.set_result(None)
+        self._disconnected_futures.clear()
+        if self._disconnect_callback:
+            self._disconnect_callback()
 
     async def connect(self) -> None:
         """Connect to the lock."""
@@ -131,7 +139,6 @@ class Lock:
             "%s: Connecting to the lock",
             self.name,
         )
-        self._disconnected_event = asyncio.Event()
         try:
             self.client = await establish_connection(
                 BleakClientWithServiceCache,
@@ -150,11 +157,15 @@ class Lock:
             self.client,
             self.name,
             self._lock,
-            self._disconnected_event,
+            self._disconnected_futures,
             self._internal_state_callback,
         )
         self.secure_session = SecureSession(
-            self.client, self.name, self._lock, self._disconnected_event, self.key_index
+            self.client,
+            self.name,
+            self._lock,
+            self._disconnected_futures,
+            self.key_index,
         )
         session = self.session
         secure_session = self.secure_session
@@ -223,7 +234,6 @@ class Lock:
         """Setup the session."""
         assert self.session is not None  # nosec
         assert self.secure_session is not None  # nosec
-        assert self._disconnected_event is not None  # nosec
         _LOGGER.debug("%s: Setting up the session", self.name)
         self.secure_session.set_key(self.key)
         handshake_keys = os.urandom(16)
@@ -387,19 +397,24 @@ class Lock:
 
         try:
             await self._shutdown_connection()
+        except BleakError:
+            _LOGGER.debug(
+                "%s: Failed to shutdown connection to lock", self.name, exc_info=True
+            )
         finally:
-            await self.client.disconnect()
+            try:
+                await self.client.disconnect()
+            except BleakError:
+                _LOGGER.debug(
+                    "%s: Failed to disconnect from lock", self.name, exc_info=True
+                )
 
     async def _shutdown_connection(self) -> None:
         """Shutdown the connection."""
         _LOGGER.debug("%s: Shutting down the connection", self.name)
         if self.session:
             await self.session.stop_notify()
-        if (
-            not self.is_secure
-            or not self.secure_session
-            or self._disconnected_event is None
-        ):
+        if not self.is_secure or not self.secure_session or self._disconnected:
             return
         cmd = self.secure_session.build_command(0x05)
         cmd[0x11] = 0x00
@@ -416,6 +431,11 @@ class Lock:
                     "%s: Failed to cleanly disconnect from lock: %s", self.name, err
                 )
             return
+        except YaleXSBLEError as err:
+            _LOGGER.debug(
+                "%s: Failed to cleanly disconnect from lock: %s", self.name, err
+            )
+            return
         if response and response[0] != 0x8B:
             _LOGGER.debug("%s: Unexpected response to DISCONNECT: %s", response.hex())
 
@@ -427,6 +447,5 @@ class Lock:
             and self.client.is_connected
             and self.session
             and self.secure_session
-            and self._disconnected_event is not None
-            and not self._disconnected_event.is_set()
+            and not self._disconnected
         )
