@@ -9,11 +9,11 @@ from collections.abc import Awaitable, Callable, Iterable
 from dataclasses import replace
 from typing import Any, TypeVar, cast
 
-import async_timeout
 from bleak.backends.device import BLEDevice
 from bleak.backends.scanner import AdvertisementData
 from bleak.exc import BleakDBusError, BleakError
 from bleak_retry_connector import BLEAK_RETRY_EXCEPTIONS, BleakNotFoundError, get_device
+from lru import LRU  # pylint: disable=no-name-in-module
 
 from .const import (
     APPLE_MFR_ID,
@@ -37,7 +37,7 @@ from .session import (
     ResponseError,
     YaleXSBLEError,
 )
-from .util import is_disconnected_error, local_name_is_unique
+from .util import asyncio_timeout, is_disconnected_error, local_name_is_unique
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -50,10 +50,16 @@ NEVER_TIME = -86400.0
 
 DEFAULT_ATTEMPTS = 4
 
+# How long to wait to disconnect after an operation
 DISCONNECT_DELAY = 5.1
 
+# How long to wait to disconnect after an operation if there is a pending update
+DISCONNECT_DELAY_PENDING_UPDATE = 12.5
+
 RESYNC_DELAY = 0.01
 
+KEEP_ALIVE_TIME = 25.0  # Lock will disconnect after 30 seconds of inactivity
+
 # Number of seconds to wait after the first connection
 # to disconnect to free up the bluetooth adapter.
 FIRST_CONNECTION_DISCONNECT_TIME = 2.1
@@ -78,7 +84,7 @@ MANUAL_UPDATE_COALESCE_SECONDS = 0.05
 POST_OPERATION_SYNC_TIME = 10.00
 
 # How long to wait if we get an update storm from the lock
-UPDATE_IN_PROGRESS_DEFER_SECONDS = 1.0
+UPDATE_IN_PROGRESS_DEFER_SECONDS = DISCONNECT_DELAY - 1
 
 RETRY_BACKOFF_EXCEPTIONS = (BleakDBusError, DisconnectedError)
 
@@ -104,6 +110,29 @@ def operation_lock(func: WrapFuncType) -> WrapFuncType:
     return cast(WrapFuncType, _async_wrap_operation_lock)
 
 
+class AuthFailureHistory:
+    """Track the number of auth failures."""
+
+    def __init__(self) -> None:
+        """Init the history."""
+        self._failures_by_mac: dict[str, int] = LRU(1024)
+
+    def auth_failed(self, mac: str) -> None:
+        """Increment the number of auth failures."""
+        self._failures_by_mac[mac] = self._failures_by_mac.get(mac, 0) + 1
+
+    def auth_success(self, mac: str) -> None:
+        """Reset the number of auth failures."""
+        self._failures_by_mac[mac] = 0
+
+    def should_raise(self, mac: str) -> bool:
+        """Return if we should raise an error."""
+        return self._failures_by_mac.get(mac, 0) >= AUTH_FAILURE_TO_START_REAUTH
+
+
+_AUTH_FAILURE_HISTORY = AuthFailureHistory()
+
+
 def retry_bluetooth_connection_error(func: WrapFuncType) -> WrapFuncType:
     """Define a wrapper to retry on bleak error.
 
@@ -122,14 +151,23 @@ def retry_bluetooth_connection_error(func: WrapFuncType) -> WrapFuncType:
             try:
                 return await func(self, *args, **kwargs)
             except AuthError:
-                self._auth_failures += 1
-                if self._auth_failures >= AUTH_FAILURE_TO_START_REAUTH:
+                _AUTH_FAILURE_HISTORY.auth_failed(self.address)
+                if _AUTH_FAILURE_HISTORY.should_raise(self.address):
                     # If the bluetooth connection drops in the middle of authentication
                     # we may see it as a failed authentication. If we see 5 failed
                     # authentications in a row we can reasonably assume that the key has
                     # changed and we should re-authenticate.
                     self._update_any_state([AuthState(successful=False)])
-                raise
+                    raise
+                _LOGGER.debug(
+                    "%s: Auth error calling %s, retrying (%s/%s)...",
+                    self.name,
+                    func,
+                    attempt,
+                    max_attempts,
+                    exc_info=True,
+                )
+                await asyncio.sleep(0.25)
             except BleakNotFoundError:
                 # The lock cannot be found so there is no
                 # point in retrying.
@@ -200,6 +238,8 @@ class PushLock:
         key_index: int | None = None,
         advertisement_data: AdvertisementData | None = None,
         idle_disconnect_delay: float = DISCONNECT_DELAY,
+        always_connected: bool = False,
+        idle_disconnect_delay_pending_update: float = DISCONNECT_DELAY_PENDING_UPDATE,
     ) -> None:
         """Init the lock watcher."""
         if local_name is None and address is None:
@@ -233,12 +273,17 @@ class PushLock:
             type[LockStatus] | type[DoorStatus] | type[BatteryState] | type[AuthState]
         ] = set()
         self._disconnect_timer: asyncio.TimerHandle | None = None
+        self._keep_alive_timer: asyncio.TimerHandle | None = None
+        self._idle_disconnect_delay_pending_update = (
+            idle_disconnect_delay_pending_update
+        )
         self._idle_disconnect_delay = idle_disconnect_delay
         self._next_disconnect_delay = idle_disconnect_delay
         self._first_update_future: asyncio.Future[None] | None = None
         self._background_tasks: set[asyncio.Task[None]] = set()
-        self._auth_failures: int = 0
         self._last_lock_operation_complete_time = NEVER_TIME
+        self._last_operation_complete_time = NEVER_TIME
+        self._always_connected = always_connected
 
     @property
     def local_name(self) -> str | None:
@@ -301,6 +346,11 @@ class PushLock:
         """Return the current BLEDevice."""
         return self._ble_device
 
+    @property
+    def is_connected(self) -> bool:
+        """Return if the lock is connected."""
+        return bool(self._client and self._client.is_connected)
+
     def set_name(self, name: str) -> None:
         """Set the name of the lock."""
         self._name = name
@@ -347,10 +397,58 @@ class PushLock:
             self.name,
             self._state_callback,
             self._lock_info,
+            self._disconnected_callback,
+        )
+
+    def _disconnected_callback(self) -> None:
+        """Handle a disconnect from the lock."""
+        _LOGGER.debug("%s: Disconnected from lock callback", self.name)
+        if self._always_connected and not _AUTH_FAILURE_HISTORY.should_raise(
+            self.address
+        ):
+            _LOGGER.debug(
+                "%s: Scheduling reconnect from disconnected callback", self.name
+            )
+            self._keep_alive()
+
+    def _keep_alive(self) -> None:
+        """Keep the lock connection alive."""
+        if not self._always_connected:
+            return
+        _LOGGER.debug("%s: Executing keep alive", self.name)
+        self._schedule_future_update(0)
+        self._schedule_next_keep_alive(KEEP_ALIVE_TIME)
+
+    def _time_since_last_operation(self) -> float:
+        """Return the time since the last operation."""
+        return time.monotonic() - self._last_operation_complete_time
+
+    def _reschedule_next_keep_alive(self) -> None:
+        """Reschedule the next keep alive."""
+        next_keep_alive_time = max(
+            0, KEEP_ALIVE_TIME - self._time_since_last_operation()
+        )
+        self._schedule_next_keep_alive(next_keep_alive_time)
+
+    def _schedule_next_keep_alive(self, delay: float) -> None:
+        """Schedule the next keep alive."""
+        self._cancel_keepalive_timer()
+        if not self._always_connected or not self._running:
+            return
+        _LOGGER.debug(
+            "%s: Scheduling next keep alive in %s seconds",
+            self.name,
+            delay,
+        )
+        self._keep_alive_timer = self.loop.call_later(
+            delay,
+            self._keep_alive,
         )
 
     def _reset_disconnect_timer(self) -> None:
         """Reset disconnect timer."""
+        if self._always_connected and self._running:
+            return
         self._cancel_disconnect_timer()
         self._expected_disconnect = False
         timeout = self._next_disconnect_delay
@@ -373,7 +471,10 @@ class PushLock:
         await self._execute_disconnect()
 
     def _disconnect_with_timer(self, timeout: float) -> None:
-        """Disconnect from device."""
+        """Disconnect from device.
+
+        This should only ever be called from _reset_disconnect_timer
+        """
         if self._operation_lock.locked():
             _LOGGER.debug("%s: Disconnect timer reset due to operation lock", self.name)
             self._reset_disconnect_timer()
@@ -395,6 +496,12 @@ class PushLock:
             self._disconnect_timer.cancel()
             self._disconnect_timer = None
 
+    def _cancel_keepalive_timer(self) -> None:
+        """Cancel keep alive timer."""
+        if self._keep_alive_timer:
+            self._keep_alive_timer.cancel()
+            self._keep_alive_timer = None
+
     async def _execute_timed_disconnect(self, timeout: float) -> None:
         """Execute timed disconnection."""
         _LOGGER.debug(
@@ -419,7 +526,9 @@ class PushLock:
     async def _execute_disconnect(self) -> None:
         """Execute disconnection."""
         async with self._connect_lock:
-            if self._disconnect_timer:  # If the timer was reset, don't disconnect
+            if (
+                self._running and self._disconnect_timer
+            ):  # If the timer was reset, don't disconnect
                 return
             client = self._client
             self._client = None
@@ -436,12 +545,14 @@ class PushLock:
                 "%s: Connection already in progress, waiting for it to complete",
                 self.name,
             )
-        if self._client and self._client.is_connected:
+        if self.is_connected:
+            assert self._client is not None  # nosec
             self._reset_disconnect_timer()
             return self._client
         async with self._connect_lock:
             # Check again while holding the lock
-            if self._client and self._client.is_connected:
+            if self.is_connected:
+                assert self._client is not None  # type: ignore[unreachable] # nosec
                 self._reset_disconnect_timer()
                 return self._client
             self._client = self._get_lock_instance()
@@ -480,6 +591,10 @@ class PushLock:
         self, op_attr: str, pending_state: LockStatus, complete_state: LockStatus
     ) -> None:
         """Execute a lock operation."""
+        if not self._running:
+            raise RuntimeError(
+                f"{self.name}: Lock operation not possible because not running"
+            )
         _LOGGER.debug("%s: Starting %s", self.name, pending_state)
         self._update_any_state([pending_state])
         self._cancel_future_update()
@@ -487,13 +602,21 @@ class PushLock:
             lock = await self._ensure_connected()
             self._cancel_future_update()
             await getattr(lock, op_attr)()
-        except Exception:
+        except Exception as ex:
             self._update_any_state([LockStatus.UNKNOWN])
+            _LOGGER.debug(
+                "%s: Failed to execute lock operation due to %s, forcing disconnect",
+                self.name,
+                ex,
+            )
             raise
         self._update_any_state([complete_state])
         _LOGGER.debug("%s: Finished %s", self.name, complete_state)
-        self._last_lock_operation_complete_time = time.monotonic()
+        now = time.monotonic()
+        self._last_lock_operation_complete_time = now
+        self._last_operation_complete_time = now
         self._reset_disconnect_timer()
+        self._reschedule_next_keep_alive()
 
     def _state_callback(
         self, states: Iterable[LockStatus | DoorStatus | BatteryState]
@@ -514,15 +637,19 @@ class PushLock:
         _LOGGER.debug("%s: State changed: %s", self.name, states)
         lock_state = self._get_current_state()
         original_lock_status = lock_state.lock
+        changes: dict[str, Any] = {}
         for state in states:
             state_type = type(state)
             self._seen_this_session.add(state_type)
             if isinstance(state, AuthState):
-                lock_state = replace(lock_state, auth=state)
+                if lock_state.auth != state:
+                    changes["auth"] = state
             elif isinstance(state, LockStatus):
-                lock_state = replace(lock_state, lock=state)
+                if lock_state.lock != state:
+                    changes["lock"] = state
             elif isinstance(state, DoorStatus):
-                lock_state = replace(lock_state, door=state)
+                if lock_state.door != state:
+                    changes["door"] = state
             elif isinstance(state, BatteryState):
                 if state.voltage <= 3.0:
                     _LOGGER.debug(
@@ -531,10 +658,15 @@ class PushLock:
                         state.voltage,
                     )
                     continue
-                lock_state = replace(lock_state, battery=state)
+                if lock_state.battery != state:
+                    changes["battery"] = state
             else:
                 raise ValueError(f"Unexpected state type: {state}")
 
+        if not changes:
+            return
+
+        lock_state = replace(lock_state, **changes)
         if (
             original_lock_status != lock_state.lock
             and (not lock_state.auth or lock_state.auth.successful)
@@ -546,10 +678,9 @@ class PushLock:
 
     async def update(self) -> None:
         """Request that status be updated."""
+        _LOGGER.debug("%s: Starting manual update", self.name)
         self._schedule_future_update_with_debounce(
-            0
-            if self._client and self._client.is_connected
-            else MANUAL_UPDATE_COALESCE_SECONDS
+            0 if self.is_connected else MANUAL_UPDATE_COALESCE_SECONDS
         )
 
     async def validate(self) -> None:
@@ -573,28 +704,41 @@ class PushLock:
         # Asking for battery first seems to be reduce the chance of the lock
         # getting into a bad state.
         state = self._get_current_state()
-        if type(BatteryState) not in self._seen_this_session:
+        made_request = False
+
+        if BatteryState not in self._seen_this_session:
+            made_request = True
             battery_state = await lock.battery()
-            self._auth_failures = 0
+            _AUTH_FAILURE_HISTORY.auth_success(self.address)
             state = replace(
                 state, battery=battery_state, auth=AuthState(successful=True)
             )
+
+        if (
+            DoorStatus not in self._seen_this_session
+            and self._lock_info
+            and self._lock_info.door_sense
+        ):
+            made_request = True
+            door_status = await lock.door_status()
+            _AUTH_FAILURE_HISTORY.auth_success(self.address)
+            state = replace(state, door=door_status, auth=AuthState(successful=True))
+
         # Only ask for the lock status if we haven't seen
         # it this session since notify callbacks will happen
         # if it changes and the extra polling can cause the lock
         # to get into a bad state.
-        if type(LockStatus) not in self._seen_this_session:
-            lock_status = await lock.lock_status()
-            self._auth_failures = 0
-            state = replace(state, lock=lock_status, auth=AuthState(successful=True))
-        if (
-            type(DoorStatus) not in self._seen_this_session
-            and self._lock_info
-            and self._lock_info.door_sense
+        #
+        # However, we always want to poll lock
+        # state to keep the connection alive if we are always connected.
+        if LockStatus not in self._seen_this_session or (
+            not made_request and self._always_connected
         ):
-            door_status = await lock.door_status()
-            self._auth_failures = 0
-            state = replace(state, door=door_status, auth=AuthState(successful=True))
+            made_request = True
+            lock_status = await lock.lock_status()
+            _AUTH_FAILURE_HISTORY.auth_success(self.address)
+            state = replace(state, lock=lock_status, auth=AuthState(successful=True))
+
         _LOGGER.debug("%s: Finished update", self.name)
         self._callback_state(state)
 
@@ -624,6 +768,9 @@ class PushLock:
             self._next_disconnect_delay = FIRST_CONNECTION_DISCONNECT_TIME
             self._reset_disconnect_timer()
 
+        if made_request:
+            self._last_operation_complete_time = time.monotonic()
+            self._reschedule_next_keep_alive()
         return state
 
     def _callback_state(self, lock_state: LockState) -> None:
@@ -721,8 +868,26 @@ class PushLock:
                 next_update,
                 scheduled_update,
             )
-        if next_update:
-            self._schedule_future_update_with_debounce(next_update)
+        if not next_update:
+            return
+        if (
+            self.is_connected
+            and self._next_disconnect_delay != FIRST_CONNECTION_DISCONNECT_TIME
+            and (
+                self._time_since_last_operation()
+                + self._idle_disconnect_delay_pending_update
+            )
+            < KEEP_ALIVE_TIME
+        ):
+            # Already connected, state will be pushed, but stay
+            # connected a bit longer to make sure we get it unless
+            # this is the first connection or deferring the update
+            # would keep the connection idle for too long and
+            # get us disconnected anyways.
+            self._next_disconnect_delay = self._idle_disconnect_delay_pending_update
+            self._reset_disconnect_timer()
+            return
+        self._schedule_future_update_with_debounce(next_update)
 
     async def start(self) -> Callable[[], None]:
         """Start watching for updates."""
@@ -735,12 +900,12 @@ class PushLock:
             self.set_ble_device(device)
             self._schedule_future_update_with_debounce(ADV_UPDATE_COALESCE_SECONDS)
 
-        def _cancel() -> None:
-            self._running = False
-            self._cancel_future_update()
-            self.background_task(self._execute_forced_disconnect("stopping"))
+        return self._cancel
 
-        return _cancel
+    def _cancel(self) -> None:
+        self._running = False
+        self._cancel_future_update()
+        self.background_task(self._execute_forced_disconnect("stopping"))
 
     def background_task(self, fut: Awaitable[Any]) -> None:
         """Execute a background task."""
@@ -755,13 +920,15 @@ class PushLock:
         if not self._first_update_future:
             raise RuntimeError("Already waited for first update")
         try:
-            async with async_timeout.timeout(timeout):
+            async with asyncio_timeout(timeout):
                 await self._first_update_future
-        except asyncio.TimeoutError:
+        except (asyncio.TimeoutError, asyncio.CancelledError) as ex:
             self._first_update_future.cancel()
             with contextlib.suppress(asyncio.CancelledError):
                 await self._first_update_future
-            raise NoAdvertisementError("No advertisement received")
+            raise NoAdvertisementError(
+                "No advertisement received before timeout"
+            ) from ex
         finally:
             self._first_update_future = None
 
@@ -773,10 +940,9 @@ class PushLock:
 
     def _schedule_future_update_with_debounce(self, seconds: float) -> None:
         """Schedule an update with a potential debounce."""
-        now = self.loop.time()
         future_update_time = seconds
         if self._cancel_deferred_update:
-            time_till_update = self._cancel_deferred_update.when() - now
+            time_till_update = self._cancel_deferred_update.when() - self.loop.time()
             if time_till_update < HK_UPDATE_COALESCE_SECONDS:
                 future_update_time = HK_UPDATE_COALESCE_SECONDS
                 _LOGGER.debug(
@@ -793,7 +959,9 @@ class PushLock:
                     time_till_update,
                 )
                 return
-            _LOGGER.debug("%s: Rescheduling update", self.name)
+            _LOGGER.debug(
+                "%s: Rescheduling update for %s", self.name, future_update_time
+            )
         self._schedule_future_update(future_update_time)
 
     def _schedule_future_update(self, future_update_time: float) -> None:
